/**
 * Copyright (c) 2010-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.orchestration.core;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

/**
 * @author gibello
 *
 */
public class MemoryDataSource implements javax.activation.DataSource {
	private byte[] data_;
	private String name_;
	private String mimeType_ = "application/octet-stream";

	/**
	 * Constructor
	 * @param data The data to keep in memory
	 * @param name Datasource name
	 */
	public MemoryDataSource(byte[] data, String name) {
		data_ = data;
		name_ = name;
	}

	/**
	 * Constructor
	 * @param data The data to keep in memory
	 * @param name Datasource name
	 * @param mimeType
	 */
	public MemoryDataSource(byte[] data, String name, String mimeType) {
		this(data, name);
		if(mimeType != null) mimeType_ = mimeType;
	}

	/**
	 * Constructor, taking a File as argument
	 * @param file The file that contains the data to keep in memory
	 * @throws IOException
	 */
	public MemoryDataSource(File file) throws IOException {
        FileInputStream fis = new FileInputStream(file);
        ByteArrayOutputStream bos = new ByteArrayOutputStream(
        		Long.valueOf(file.length()).intValue());
        copyStream(fis, bos);
        fis.close();
        bos.close();

        name_ = file.getName();
        data_ = bos.toByteArray();
    }

	/**
	 * @see javax.activation.DataSource#getContentType()
	 */
	public String getContentType() {
		return mimeType_;
	}

	/**
	 * @see javax.activation.DataSource#getInputStream()
	 */
	public InputStream getInputStream() throws IOException {
		return new ByteArrayInputStream(data_);
	}

	/**
	 * @see javax.activation.DataSource#getName()
	 */
	public String getName() {
		return name_;
	}

	/**
	 * @see javax.activation.DataSource#getOutputStream()
	 */
    public OutputStream getOutputStream() throws IOException {
        OutputStream os = new ByteArrayOutputStream();
        copyStream(getInputStream(), os);
        return os;
    }
	
    /**
     * Copies the content from in into os.
     * <p>
     * Neither <i>in</i> nor <i>os</i> are closed by this method.<br />
     * They must be explicitly closed after this method is called.
     * </p>
     * 
     * @param in
     * @param os
     * @throws IOException
     */
    private final static void copyStream(InputStream in, OutputStream os) throws IOException {
        byte[] buf = new byte[1024];
        int len;
        while((len = in.read( buf )) > 0) {
            os.write( buf, 0, len );
        }
    }
}
