/**
 * Copyright (c) 2010-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.orchestration.core;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.FileReader;
import java.io.InputStream;
import java.io.StringReader;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.zip.ZipInputStream;

import javax.activation.DataHandler;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.URIResolver;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.jbidescriptor.generated.Provides;
import org.ow2.petals.component.framework.message.ExchangeImpl;
import org.ow2.petals.component.framework.util.ExchangeUtil;
import org.ow2.petals.component.framework.util.SourceUtil;
import org.w3c.dom.Document;
import org.xml.sax.InputSource;

import com.ebmwebsourcing.easycommons.xml.XMLHelper;

public class OrchestrationUtils {

	public static String getEndpointName(Consumes ep) {
		QName ifn = ep.getInterfaceName();
		QName svn = ep.getServiceName();

		String ifns = ifn.getNamespaceURI();
		String svns = svn.getNamespaceURI();

		StringBuffer name = new StringBuffer((ifns == null ? "" : "{" + ifns + "}") + ifn.getLocalPart());
		name.append("|" + (svns == null || svns.equals(ifns) ? "" : "{" + svns + "}") + svn.getLocalPart());
		name.append("|" + ep.getEndpointName());
		return name.toString();
	}
	public static String getEndpointName(Provides ep) {
		QName ifn = ep.getInterfaceName();
		QName svn = ep.getServiceName();

		String ifns = ifn.getNamespaceURI();
		String svns = svn.getNamespaceURI();

		StringBuffer name = new StringBuffer((ifns == null ? "" : "{" + ifns + "}") + ifn.getLocalPart());
		name.append("|" + (svns == null || svns.equals(ifns) ? "" : "{" + svns + "}") + svn.getLocalPart());
		name.append("|" + ep.getEndpointName());
		return name.toString();
	}
	public static String getEndpointName(ServiceEndpoint ep) {
		QName ifs[] = ep.getInterfaces();
        QName ifn = ifs[0];
		QName svn = ep.getServiceName();
		
		String ifns = ifn.getNamespaceURI();
		String svns = svn.getNamespaceURI();
		
		StringBuffer name = new StringBuffer((ifns == null ? "" : "{" + ifns + "}") + ifn.getLocalPart());
		name.append("|" + (svns == null || svns.equals(ifns) ? "" : "{" + svns + "}") + svn.getLocalPart());
		name.append("|" + ep.getEndpointName());
		return name.toString();
	}
	
	public static Consumes getConsumesEndpoint(String epr) throws Exception {
		if(epr == null) return null;
		epr = epr.trim();
		StringTokenizer st = new StringTokenizer(epr, "|");
		if(! st.hasMoreTokens()) return null;
		String interfac = st.nextToken();
		String service = (st.hasMoreTokens() ? st.nextToken() : null);
		String endpoint = (st.hasMoreTokens() ? st.nextToken() : null);
		
		if(epr.startsWith("|")) {
			endpoint = service;
			service = interfac;
			interfac = null;
		}
		
		String interfaceNS = extractNS(interfac, null);
		String serviceNS = extractNS(service, null);

		if(interfaceNS == null && serviceNS == null)
			throw new Exception("No namespace specified for endpoint " + epr);
		else if(interfaceNS == null) interfaceNS = serviceNS;
		else if(serviceNS == null) serviceNS = interfaceNS;
		
		Consumes ep = new Consumes();
		if(interfac != null) ep.setInterfaceName(new QName(interfaceNS, extractLocalPart(interfac)));
		if(service != null) ep.setServiceName(new QName(serviceNS, extractLocalPart(service)));
		if(endpoint != null) ep.setEndpointName(endpoint);
		
		//System.err.println("interface=" + interfac + ",service=" + service + ",endpoint=" +  endpoint);
		return ep;
	}
	
	public static String extractNS(String qname, String dflt) {
		if(qname == null) return dflt;
		int start = qname.indexOf('{');
		if(start < 0) return dflt;
		int end = qname.indexOf('}');
		if(end <= start+1) return dflt;
		return qname.substring(start+1, end);
	}
	public static String extractLocalPart(String qname) {
		if(qname == null) return null;
		int start = qname.indexOf('}');
		if(start < 0) return qname;
		else if(qname.length() > start+1) return qname.substring(start+1);
		else return null;	
	}
	
	/**
	 * Switch attachment to payload (the previous payload gets lost, and the corresponding attachment is removed).
	 * When no attachment is eligible, nothing is done.
	 * @param exchange The exchange.
	 * @param attname The attachment name (null means 1st attachment).
	 */
	public static void attachmentToPayload(Exchange exchange, String attname) {
		try {
			boolean isOut = exchange.isOutMessage();

			DataHandler attachment = null;
			Set<String> attnames = (isOut ? exchange.getOutMessageAttachmentNames() : exchange.getInMessageAttachmentNames());
			if(! attnames.isEmpty()) {
				String name = null;
				if(attname == null) {
					name = attnames.iterator().next();
				} else {
					for(String n : attnames) {
						if(attname.equals(n)) {
							name = n;
							break;
						}
					}
				}
				
				if(name != null) {
					attachment = (isOut ? exchange.getOutMessageAttachment(name) : exchange.getInMessageAttachment(name));
					attnames.remove(name);

					//Set atts = (isOut ? exchange.getOutMessageAttachments() : exchange.getInMessageAttachments());
					//System.out.println(atts.remove(attachment));
					if(isOut) exchange.setOutMessageContent(attachment.getDataSource().getInputStream());
					else exchange.setInMessageContent(attachment.getDataSource().getInputStream());
				}
			}

		} catch(Exception e) {
			e.printStackTrace();
		}
	}
	
	public static void unzipAttachment(Exchange exchange, String attname) throws Exception {
		boolean isOut = exchange.isOutMessage();
		DataHandler attachment = (isOut ? exchange.getOutMessageAttachment(attname) : exchange.getInMessageAttachment(attname));
		DataHandler result = unzip(attachment);
		if(result != null) {
			if(isOut) exchange.setOutMessageAttachment(attname, result);
			else exchange.setInMessageAttachment(attname, result);
		}
	}
	
	public static void unzipAttachments(Exchange exchange) throws Exception {
		boolean isOut = exchange.isOutMessage();
		Set<String> names = (isOut ? exchange.getOutMessageAttachmentNames() : exchange.getInMessageAttachmentNames());
		for(String name : names) {
			unzipAttachment(exchange, name);
		}
	}
	
	public static void transform(Exchange exchange, String xslt, URIResolver r, String attname) throws Exception {
		TransformerFactory tFactory = TransformerFactory.newInstance();
	    Transformer transformer = tFactory.newTransformer(new StreamSource(new FileReader(xslt)));
	    if(r != null) transformer.setURIResolver(r);
	    boolean isOut = exchange.isOutMessage();
	    Source in = (isOut ? exchange.getOutMessageContentAsSource() : exchange.getInMessageContentAsSource());
	    ByteArrayOutputStream out = new ByteArrayOutputStream();
	    transformer.transform(in, new StreamResult(out));
	    //System.out.println(out);
	    
	    if(attname == null) {
	    	if(isOut) {
	    		exchange.setOutMessageContent(new ByteArrayInputStream(out.toByteArray()));
	    	} else {
	    		exchange.setInMessageContent(new ByteArrayInputStream(out.toByteArray()));
	    	}
	    } else {
	    	StreamSource content = new StreamSource(new StringReader("<attached-files><file-name>"
                    + attname + "</file-name></attached-files>"));
	    	DataHandler attachment = new DataHandler(new MemoryDataSource(out.toByteArray(), attname, "application/octet-stream")); 
	    	if(isOut) {
	    		exchange.setOutMessageContent(content);
	    		exchange.addOutMessageAttachment(attname, attachment);
	    	} else {
	    		exchange.setInMessageContent(content);
	    		exchange.addInMessageAttachment(attname, attachment);
	    	}
	    }
	}

	/**
	 * Unzip DataHandler content.
	 * @param dataHandler The DataHandler to unzip
	 * @return A new DataHandler, with unzipped content
	 */
    public static DataHandler unzip(DataHandler dataHandler) {
        ByteArrayOutputStream out = new ByteArrayOutputStream();
        try (ZipInputStream zIn = new ZipInputStream(dataHandler.getInputStream())) {
            zIn.getNextEntry();
            byte[] buffer = new byte[1024];
            int count = 0;
            do {
                count = zIn.read(buffer, 0, 1024);
                if (count > 0)
                    out.write(buffer, 0, count);
            } while (count != -1);
            return new DataHandler(
                    new MemoryDataSource(out.toByteArray(), dataHandler.getName(), dataHandler.getContentType()));
        } catch (Exception e) {
            // e.printStackTrace(System.err);
            return null;
        }
    }

	/**
	 * Convert Source content to String
	 * @param source The Source to convert
	 * @return The Source content, as a String
	 * @throws Exception
	 */
	public static String sourceToString(Source source) throws Exception {
		String ret = null;
		if(source instanceof StreamSource) {
			InputStream is = ((StreamSource)source).getInputStream();
			int len = is.available();
			byte[] bytes = new byte[len];
			int count = is.read(bytes,0,len);
			if(count > 0) ret = new String(bytes, 0, count);
			is.close();
		} else {
			ret = XMLHelper.createStringFromDOMDocument(((DOMSource)source).getNode());
		}
		return ret;
	}
	
	public static Document stringToDocument(String xml) throws Exception {
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		InputSource source = new InputSource(new StringReader(xml));
		return factory.newDocumentBuilder().parse(source);
	}
	
	/**
	 * Copied from org.ow2.petals.component.framework.util.ExchangeUtilImpl
	 * from petals-cdk-core 5.1.1
	 */
    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.util.ExchangeUtil#copyExchange
     * (javax.jbi.messaging.MessageExchangeFactory,
     * org.ow2.petals.component.framework.api.message.Exchange)
     */
    public static Exchange copyExchange(final MessageExchangeFactory mef, final Exchange source) throws MessagingException,
            PEtALSCDKException {

        final MessageExchange msgEx = mef.createExchange(source.getPattern());
        final Exchange target = new ExchangeImpl(msgEx);

        // Set all values in concrete message exchange to fake message exchange
        if (source.getEndpoint() != null) {
            target.setEndpoint(source.getEndpoint());
        }
        target.setInterfaceName(source.getInterfaceName());
        target.setService(source.getService());
        target.setOperation(source.getOperation());
        target.setStatus(source.getStatus());

        if (source.getError() != null) {
            target.setError(source.getError());
        }

        if (source.getFault() != null) {
            target.setFault(source.getFault());
        }

        if (source.getInMessage() != null && source.getInMessage().getContent() != null) {
            copyNormalizedMessage(source.getInMessage(), target.getInMessage());
        }

        if (source.getOutMessage() != null && source.getOutMessage().getContent() != null) {
            copyNormalizedMessage(source.getOutMessage(), target.getOutMessage());
        }
        ExchangeUtil.copyExchangeProperties(source, target);
        return target;
    }

    /*
     * (non-Javadoc)
     * @see org.ow2.petals.component.framework.api.util.IExchangeUtil#
     * copyNormalizedMessage(javax.jbi.messaging.NormalizedMessage, javax.jbi.messaging.NormalizedMessage)
     */
    @SuppressWarnings("unchecked")
    public static void copyNormalizedMessage(final NormalizedMessage originalNm, final NormalizedMessage copyNm)
            throws MessagingException, PEtALSCDKException {
        // setting content, with hardcopy (because if semantic routing...)
        final Document contentDoc = SourceUtil.createDocument(originalNm.getContent());
        final Source source = SourceUtil.createSource(contentDoc);
        copyNm.setContent(source);
        // a copy is not a consume : set again the original content read
        originalNm.setContent(source);
        // setting properties
        final Set<String> propertyNames = originalNm.getPropertyNames();
        for (final String propertyName : propertyNames) {
            copyNm.setProperty(propertyName, originalNm.getProperty(propertyName));
        }

        // setting attachments if any
        final Set<String> inAttachmentNames = originalNm.getAttachmentNames();
        if (inAttachmentNames != null) {
            for (final String attachmentName : inAttachmentNames) {
                final DataHandler attDh = originalNm.getAttachment(attachmentName);
                copyNm.addAttachment(attachmentName, attDh);
            }
        }

        // setting security context
        copyNm.setSecuritySubject(originalNm.getSecuritySubject());
    }
}
